#include "main.h"

static char szFilesNeeded[4][16]=
{
   "mame.rms",
   "mameinfo.lst",
   "mame.lst",
   "arcadeos.cfg"
};

static char yesno[3][16]=
{
   DONT_CARE,
   "no",
   "yes"
};

static char szDC[]=DONT_CARE;

pBYTE pInfo;
pBYTE pMAME;

CLIST clistdefs[MAX_CLIST_DEFS];
int NoDefs=0;
int bReportOnly=FALSE;
int main(int argc,char **argv)
{
   FILE *fCl;
//check we've got enough arguments
   printf("\nCLIST Version 1.0\n");
   if(helpwanted(argc,argv))
   {
      printf("Custom list generator for ArcadeOS\n\n");
      printf("Usage:\nCLIST <clist file name> [/r] \n\nSee clist.cl for examples of custom list file contents\n");
      printf("\n/r indicates that CLIST will just report which games match the criteria \nbut will *NOT* copy them into a custom list\n");
      return 0;
   }
   if(argc==3)
   {
      bReportOnly=TRUE;
      printf("Report Mode - CLIST will not copy games into custom lists,\njust report which ones match\n");
   }

//check all the files we need are present
   if(!checkclistfiles())
      return 0;
//check it's the right RMS version
   if(!checkrms())
   {
      printf("\nIncorrect .RMS file format for CLIST\n\n");
      return 0;
   }
//read in the MAME info
   if(!getinfo())
   {
      printf("\nUnable to read in MAMEINFO.LST\n\n");
      return 0;
   }
   printf("\nUsing custom list file \"%s\"\n",argv[1]);
   fCl=fopen(argv[1],"rt");
   if(!fCl)
   {
      printf("\nUnable to open file \"%s\" CLIST Quitting\n\n",argv[1]);
      free(pInfo);
      free(pMAME);
      return 0;
   }

   if(!parseinfo(fCl))
   {
      free(pInfo);
      free(pMAME);
      return 0;
   }

   if(NoDefs==0)
   {
      printf("\nNo custom list definitions found in \"%s\"\n\n",argv[1]);
      fclose(fCl);
      free(pInfo);
      free(pMAME);
      return 0;
   }
   else
   {
      if(NoDefs==1)
         printf("\nFound 1 definition in \"%s\"\n",argv[1]);
      else
         printf("\nFound %d definitions in \"%s\"\n",NoDefs,argv[1]);
   }

   createclists();
   fclose(fCl);

   free(pInfo);
   free(pMAME);
   return 0;
}


int checkclistfiles(void)
{
   int i;
   for(i=0;i<4;i++)
   {
      if(!checkfile(szFilesNeeded[i]))
      {
         printf("\nCLIST requires the file \"%s\" to be present\nPlease run CLIST in your ArcadeOS directory\n\n",szFilesNeeded[i]);
         return FALSE;
      }
   }
   return TRUE;
}
int checkfile(char *szfile)
{
	FILE	*fchk;

   fchk=fopen(szfile,"rb");
   if(!fchk)
      return FALSE;
   fclose(fchk);
	return TRUE;

}

int checkrms(void)
{
   FILE *fchk;
   char szIn[RMS_HEAD_SIZE];

   fchk=fopen("mame.rms","rb");
   if(!fchk)
      return FALSE;
   fread(szIn,1,RMS_HEAD_SIZE,fchk);
   if(memcmp(szIn,RMSFORMAT,RMS_HEAD_SIZE))
   {
      fclose(fchk);
      return FALSE;
   }
   fclose(fchk);
   return TRUE;
}

int getinfo(void)
{
   FILE *fInfo;
   FILE *fMAME;


   fInfo=fopen("mameinfo.lst","rb");
   if(!fInfo)
      return FALSE;

   fMAME=fopen("mame.lst","rb");
   if(!fMAME)
   {
      fclose(fInfo);
      return FALSE;
   }
   pInfo=malloc(MAX_INFO_SIZE);
   if(!pInfo)
   {
      fclose(fInfo);
      return FALSE;
   }
   pMAME=malloc(MAX_INFO_SIZE);
   if(!pMAME)
   {
      fclose(fInfo);
      free(pInfo);
      return FALSE;

   }
   fread(pInfo,1,MAX_INFO_SIZE,fInfo);
   fclose(fInfo);
   fread(pMAME,1,MAX_INFO_SIZE,fMAME);
   fclose(fMAME);
   return TRUE;

}


int parseinfo(FILE *fCL)
{
   char buffer[1024];
   int found=FALSE;
   int curdef=-1;

   while(tidyget(buffer,1023,fCL))
   {
      if(!strstr(buffer,"#"))
      {
         if(strstr(buffer,"customlist"))
         {
            curdef++;
            clistdefs[curdef].nClist=getval(buffer);
            clistdefs[curdef].minyear=0;
            clistdefs[curdef].maxyear=9999;
            clistdefs[curdef].minx=0;
            clistdefs[curdef].maxx=9999;
            clistdefs[curdef].miny=0;
            clistdefs[curdef].maxy=9999;
            clistdefs[curdef].vector=-1;
            clistdefs[curdef].szMan[0]=0;
            clistdefs[curdef].szName[0]=0;
            clistdefs[curdef].vertical=-1;
            clistdefs[curdef].clone=-1;
            clistdefs[curdef].szCloneName[0]=0;
            clistdefs[curdef].szStartName[0]=0;
            clistdefs[curdef].neogeo=-1;
            clistdefs[curdef].players=-1;
            clistdefs[curdef].depth=-1;
            clistdefs[curdef].gameokay=-1;
            clistdefs[curdef].minfreq=0;
            clistdefs[curdef].maxfreq=9999;
            clistdefs[curdef].mincolors=0;
            clistdefs[curdef].maxcolors=9999;
            found=TRUE;
         }
         if(found)
         {
            if(findident(buffer,"man"))
               getstring(buffer,clistdefs[curdef].szMan);
            if(findident(buffer,"year"))
               getrange(buffer,&clistdefs[curdef].minyear,&clistdefs[curdef].maxyear);
            if(findident(buffer,"xres"))
               getrange(buffer,&clistdefs[curdef].minx,&clistdefs[curdef].maxx);
            if(findident(buffer,"yres"))
               getrange(buffer,&clistdefs[curdef].miny,&clistdefs[curdef].maxy);

            if(findident(buffer,"name"))
               getstring(buffer,clistdefs[curdef].szName);
            if(findident(buffer,"vertical"))
               clistdefs[curdef].vertical=getyesno(buffer);
            if(findident(buffer,"vector"))
               clistdefs[curdef].vector=getyesno(buffer);
            if(findident(buffer,"clone"))
               clistdefs[curdef].clone=getyesno(buffer);
            if(findident(buffer,"neogeo"))
               clistdefs[curdef].neogeo=getyesno(buffer);

            if(findident(buffer,"iscloneof"))
               getstring(buffer,clistdefs[curdef].szCloneName);
            if(findident(buffer,"players"))
               clistdefs[curdef].players=getval(buffer);
            if(findident(buffer,"depth"))
               clistdefs[curdef].depth=getval(buffer);
            if(findident(buffer,"startname"))
               getstring(buffer,clistdefs[curdef].szStartName);

            if(findident(buffer,"okay"))
               clistdefs[curdef].gameokay=getyesno(buffer);
            if(findident(buffer,"refresh"))
               getrange(buffer,&clistdefs[curdef].minfreq,&clistdefs[curdef].maxfreq);
            if(findident(buffer,"colors"))
               getrange(buffer,&clistdefs[curdef].mincolors,&clistdefs[curdef].maxcolors);
         }
      }
   }
   NoDefs=curdef+1;
   return TRUE;
}

int findident(char *buffer,char *ident)
{
   char *pWork;
   pWork=trimspaces(buffer);
   if(strstr(pWork,ident)==pWork)
      return TRUE;
   return FALSE;
}

char *tidyget(char *inbuff,int len,FILE *f)
{

	char *ret;
	*inbuff=0;

	ret=NULL;
	if(f)
	{
		ret=fgets(inbuff,len,f);
		if(ret)
		{
			inbuff[len]=0;
			if(strlen(inbuff))
			{
				if(inbuff[strlen(inbuff)-1]<32)
					inbuff[strlen(inbuff)-1]=0;
			}
		}
	}
	return ret;
}

void getstring(char *buffer,char *out)
{
   char *pWork;
   *out=0;
   pWork=getequalval(buffer);
   if(pWork)
      strcpy(out,pWork);
}

int getyesno(char *buffer)
{
   int ret=-1;
   char *pWork;
   pWork=getequalval(buffer);
   if(pWork)
   {
      if(!stricmp(pWork,"yes"))
         ret=TRUE;
      if(!stricmp(pWork,"no"))
         ret=FALSE;
   }
   return ret;
}
int getval(char *buffer)
{
   int ret=-1;
   char *pWork;
   pWork=getequalval(buffer);
   if(pWork&&isdigit(*pWork))
      ret=atoi(pWork);
   return ret;
}

void getrange(char *buffer,int *pmin,int *pmax)
{
   char *pWork;
   char *pWork2;
   char szWork[256];
   pWork=getequalval(buffer);
   if(isdigit(*pWork))
   {
//either equal to or range
      pWork2=strstr(pWork,"!");
      if(pWork2)
      {
//range
         *pWork2=0;
         pWork2++;
         strcpy(szWork,pWork2);
         pWork2=trimspaces(szWork);
         *pmin=atoi(pWork);
         *pmax=atoi(pWork2);
      }
      else
      {
//one value
         *pmin=atoi(pWork);
         *pmax=atoi(pWork);
      }
   }
   else
   {
//greate than or less than
      if(*pWork=='<')
      {
         pWork++;
         *pmax=atoi(pWork)-1;
         *pmin=0;
      }
      if(*pWork=='>')
      {
         pWork++;
         *pmax=9999;
         *pmin=atoi(pWork)+1;
      }
   }
}
char *trimspaces(char *in)
{
	char *out;
	char *pWork;

	out=in;

	while(isspace(*out))
		out++;
	if(strlen(out))
	{
		pWork=out+strlen(out)-1;
		while(isspace(*pWork)&&pWork>in)
			pWork--;
		pWork++;
		*pWork=0;
	}

	return out;
}

char *getequalval(char *buffer)
{
   char *pRet;
   pRet=strstr(buffer,"=");
   if(pRet)
   {
      pRet++;
      pRet=trimspaces(pRet);
   }
   return pRet;
}

void createclists(void)
{
   int n,nogames;
   FILE *frms;
   FILE *fout;
   FILE *fcfg;
   FILE *fcfgout;
   char *pWork;
   MAMEGAME game;
   MAMEDESC gameinfo;
   char szWork[256];
   char buffer[1024];

   fcfg=fopen("arcadeos.cfg","rt");
   if(!fcfg)
      return;
   fcfgout=fopen("cfgout$.cfg","wt");
   if(!fcfgout)
   {
      fclose(fcfg);
      return;
   }

   while(tidyget(buffer,1023,fcfg))
   {
      if(!strstr(buffer,"#"))
      {
         for(n=0;n<NoDefs;n++)
         {
            sprintf(szWork,"clist%d",clistdefs[n].nClist);
            if(strstr(buffer,szWork))
            {
               pWork=getequalval(buffer);
               if(pWork&&*clistdefs[n].szName)
                  strcpy(pWork,clistdefs[n].szName);
            }
         }
      }
      fprintf(fcfgout,"%s\n",buffer);
   }
   fclose(fcfg);
   fclose(fcfgout);
   unlink("arcadeos.cfg");
   rename("cfgout$.cfg","arcadeos.cfg");

   for(n=0;n<NoDefs;n++)
   {
      frms=fopen("mame.rms","rb");
      if(!frms)
      {
         unlink("cfgout$.cfg");
         return;
      }

      fout=fopen("out.rms","wb");
      if(!fout)
      {
         fclose(frms);
         unlink("cfgout$.cfg");
         return;
      }

      printf("\n-----------------------------------------------\n");
      printf("                 Definition %d\n",n+1);
      printf("-----------------------------------------------\n");
      printf("Custom List No    : %d\n",clistdefs[n].nClist);
      printf("Name of List      : %s\n",clistdefs[n].szName);
      printf("Min Year          : %s\n",disprange(clistdefs[n].minyear));
      printf("Max Year          : %s\n",disprange(clistdefs[n].maxyear));
      printf("Min X Res         : %s\n",disprange(clistdefs[n].minx));
      printf("Max X Res         : %s\n",disprange(clistdefs[n].maxx));
      printf("Min Y Res         : %s\n",disprange(clistdefs[n].miny));
      printf("Max Y Res         : %s\n",disprange(clistdefs[n].maxy));
      printf("Name Starts With  : %s\n",dispstr(clistdefs[n].szStartName));
      printf("Manufacturer      : %s\n",dispstr(clistdefs[n].szMan));
      printf("Vertical ?        : %s\n",yesno[clistdefs[n].vertical+1]);
      printf("Vector ?          : %s\n",yesno[clistdefs[n].vector+1]);
      printf("Clone ?           : %s\n",yesno[clistdefs[n].clone+1]);
      printf("NeoGeo ?          : %s\n",yesno[clistdefs[n].neogeo+1]);
      printf("Is Clone of       : %s\n",dispstr(clistdefs[n].szCloneName));
      printf("Players           : %s\n",dispno(clistdefs[n].players));
      printf("Colour Depth      : %s\n",dispno(clistdefs[n].depth));
      printf("Game runs OK      : %s\n",yesno[clistdefs[n].gameokay+1]);
      printf("Min Refresh Rate  : %s\n",disprange(clistdefs[n].minfreq));
      printf("Max Refresh Rate  : %s\n",disprange(clistdefs[n].maxfreq));
      printf("Min No.of Colours : %s\n",disprange(clistdefs[n].mincolors));
      printf("Max No.of Colours : %s\n",disprange(clistdefs[n].maxcolors));
      printf("-----------------------------------------------\n");
      nogames=0;
      fseek(frms,0,SEEK_SET);
      fread(szWork,1,RMS_HEAD_SIZE,frms);
      fwrite(szWork,1,RMS_HEAD_SIZE,fout);
      while(fread(&game,1,sizeof(MAMEGAME),frms))
      {
//get info for this game
         if(getgameinfo(&game,&gameinfo))
//set custom list
            setcustom(&game,&gameinfo,&clistdefs[n],&nogames);
//write changes
         fwrite(&game,1,sizeof(MAMEGAME),fout);
      }
      printf("------------\n");
      printf("%d games\n",nogames);
      printf("------------\n\n");
      fclose(fout);
      fclose(frms);
//copy file across
      unlink("mame.rms");
      rename("out.rms","mame.rms");
   }
}

char *dispno(int in)
{
   static char szOut[16];
   if(in==-1)
      return szDC;
   else
      sprintf(szOut,"%d",in);
   return szOut;
}
int getgameinfo(pMAMEGAME pGame,pMAMEDESC pGameInfo)
{
   char  szWork[256];
   char  szWork2[256];
   char  *pWork;
   char  *pWork2;
   char  cwork;
   int   bFound=FALSE;

   sprintf(szWork,"^%s{",pGame->GameName);
   pWork=strstr((char *)pInfo,szWork);

   *pGameInfo->szCloneOf=0;
   *pGameInfo->szDescription=0;
   pGameInfo->nClone=FALSE;
   pGameInfo->nRotated=FALSE;
   pGameInfo->nVector=FALSE;

   if(pWork)
   {
      bFound=TRUE;
//get the bit stuff
      pWork2=strstr(pWork,"@");
      if(!pWork2)
         return FALSE;

      pWork2++;
      cwork=*pWork2;
      pGameInfo->nOkay=cwork&1;
      pGameInfo->nVector=cwork&2;
      pGameInfo->nHiScoreOK=cwork&4;
      pWork2++;
      pGameInfo->nFreq=(int)*pWork2;
      pWork2++;
      pGameInfo->nPlayers=(int)*pWork2;
      pWork2++;
      pGameInfo->nCoins=(int)*pWork2;
      pWork2++;
      pGameInfo->nColorDepth=(int)*pWork2;
      pWork2+=2;
      getitem(szWork,pWork2,0,127,'{');
      pGameInfo->nColors=atoi(szWork);
      pWork2=strstr(pWork2,"{");
      pWork2++;

      getitem(pGameInfo->szResolution,pWork2,0,15,'{');
      getitem(pGameInfo->szControl,pWork2,1,255,'{');
      getitem(pGameInfo->szYear,pWork2,2,7,'{');
      getitem(pGameInfo->szMan,pWork2,3,255,'{');
   }

   sprintf(szWork,"^%s{",pGame->GameName);
   pWork=strstr((char *)pMAME,szWork);

   if(pWork)
   {
      getitem(pGameInfo->szDescription,pWork,1,255,'{');
      getitem(pGameInfo->szCloneOf,pWork,2,255,'{');

      if(*pGameInfo->szCloneOf&&stricmp(pGameInfo->szCloneOf,NEOGEO))
         pGameInfo->nClone=TRUE;
      pWork2=getitem(szWork2,pWork,3,255,'{');
      if(pWork2)
      {
         pWork2++;
         pGameInfo->nRotated=*pWork2&1;
         pGameInfo->nVector=(*pWork2&4)>>2;
      }
   }

   return bFound;
}

void setcustom(pMAMEGAME pGame,pMAMEDESC pGameInfo,pCLIST pClistDef,int *nogames)
{
   int nYear;
   int nX,nY;
   char szWork1[256];
   char szWork2[256];
//check name start
   if(*pClistDef->szStartName)
   {
      strcpy(szWork1,pClistDef->szStartName);
      strcpy(szWork2,pGameInfo->szDescription);
      strupr(szWork1);
      strupr(szWork2);
      if(strstr(szWork2,szWork1)!=szWork2)
         return;
   }
//check manufacture
   if(*pClistDef->szMan)
      if(stricmp(pClistDef->szMan,pGameInfo->szMan))
         return;
//check clone of
   if(*pClistDef->szCloneName)
      if(stricmp(pClistDef->szCloneName,pGameInfo->szCloneOf))
         return;

//check vertical
   if(pClistDef->vertical!=-1)
      if(pClistDef->vertical!=pGameInfo->nRotated)
         return;
//check vector
   if(pClistDef->vector!=-1)
      if(pClistDef->vector!=pGameInfo->nVector)
         return;
//check clone
   if(pClistDef->clone!=-1)
      if(pClistDef->clone!=pGameInfo->nClone)
         return;
//check players
   if(pClistDef->players!=-1)
      if(pClistDef->players!=pGameInfo->nPlayers)
         return;
//check colour depth
   if(pClistDef->depth!=-1)
      if(pClistDef->depth!=pGameInfo->nColorDepth)
         return;
//check neogro
   if(pClistDef->neogeo!=-1)
      if((pClistDef->neogeo&&stricmp(pGameInfo->szCloneOf,NEOGEO))
            ||(!pClistDef->neogeo&&!stricmp(pGameInfo->szCloneOf,NEOGEO)))
            return;
//check resoltion
   if(!pGameInfo->nVector)
   {
      getres(pGameInfo->szResolution,&nX,&nY);
      if(nX<pClistDef->minx||nX>pClistDef->maxx)
         return;
      if(nY<pClistDef->miny||nY>pClistDef->maxy)
         return;
   }
//check game okay
   if(pClistDef->gameokay!=-1)
      if(pClistDef->gameokay!=pGameInfo->nOkay)
         return;
//refresh rate
   if(pGameInfo->nFreq<pClistDef->minfreq||pGameInfo->nFreq>pClistDef->maxfreq)
      return;
//colours
   if(pGameInfo->nColors<pClistDef->mincolors||pGameInfo->nColors>pClistDef->maxcolors)
      return;

//finally the year
   nYear=atoi(pGameInfo->szYear);
   if(nYear<pClistDef->minyear||nYear>pClistDef->maxyear)
      return;
   *nogames=*nogames+1;
//set the custom list
   if(bReportOnly)
      printf("Report Mode : %s\n",pGameInfo->szDescription);
   else
   {
      printf("%s\n",pGameInfo->szDescription);
      pGame->Orientation&=7;
      pGame->Orientation|=(pClistDef->nClist<<4);
   }
}


char *getitem(char *out,char *in,int nindex,int bsize,char sep)
{
	int n;
	char *pWork;
	char *pWork2;
	char szSep[2];
	szSep[0]=sep;
	szSep[1]=0;

	pWork=in;

	for(n=0;n<nindex;n++)
	{
      pWork=strstr(pWork,szSep);
		if(pWork)
			pWork++;
	}
	if(pWork)
	{
		strncpy(out,pWork,bsize);
	}
	else
	{
		strncpy(out,in,bsize);
	}

	n=bsize;
	while(n>0&&(!out[n]||isspace(out[n])))
	{
		out[n]=0;
		n--;
	}

	pWork=out;
	while(isspace(*pWork))
		pWork++;
	pWork2=strstr(out,szSep);

	if(pWork2)
		*pWork2=0;

	return pWork;
}

void getres(char *szRes,int *pX,int *pY)
{
   char szWork[256];
   char *pWork;
   *pX=*pY=0;

   strcpy(szWork,szRes);
   pWork=strstr(szWork,"x");
   if(pWork)
   {
      *pWork=0;
      pWork++;
      *pX=atoi(szWork);
      *pY=atoi(pWork);
   }
}

char *dispstr(char *szIn)
{
   if(!*szIn)
      return szDC;
   return szIn;
}

char *disprange(int in)
{
   static char szOut[16];
   if(in==0||in==9999)
      return szDC;
   sprintf(szOut,"%d",in);
   return szOut;
}

int helpwanted(int argc,char **argv)
{
   if(argc<2)
      return TRUE;
   if(!stricmp(argv[1],"/?")||!stricmp(argv[1],"-?")||
         !stricmp(argv[1],"/h")||!stricmp(argv[1],"-h"))
      return TRUE;

   return FALSE;
}


